// @flow
import * as I from 'immutable'
import {compose, connect, setDisplayName, type TypedState} from '../util/container'
import Files from '.'
import * as Types from '../constants/types/fs'
import * as Constants from '../constants/fs'
import {
  sortRowItems,
  type SortableStillRowItem,
  type SortableEditingRowItem,
  type SortableUploadingRowItem,
  type SortableRowItem,
} from './utils/sort'
import SecurityPrefsPromptingHoc from './common/security-prefs-prompting-hoc'
import FilesLoadingHoc from './files-loading-hoc'

const mapStateToProps = (state: TypedState, {path}) => {
  const itemDetail = state.fs.pathItems.get(path, Constants.makeUnknownPathItem())
  const itemChildren = itemDetail.type === 'folder' ? itemDetail.get('children', I.Set()) : I.Set()
  const itemFavoriteChildren =
    itemDetail.type === 'folder' ? itemDetail.get('favoriteChildren', I.Set()) : I.Set()
  const _username = state.config.username || undefined
  const resetParticipants =
    itemDetail.type === 'folder' && !!itemDetail.tlfMeta && itemDetail.tlfMeta.resetParticipants.length > 0
      ? itemDetail.tlfMeta.resetParticipants.map(i => i.username)
      : []
  const isUserReset = resetParticipants.includes(_username)
  const _downloads = state.fs.downloads
  return {
    _itemChildren: itemChildren,
    _itemFavoriteChildren: itemFavoriteChildren,
    _pathItems: state.fs.pathItems,
    _edits: state.fs.edits,
    _sortSetting: state.fs.pathUserSettings.get(path, Constants.makePathUserSetting()).get('sort'),
    _username,
    isUserReset,
    resetParticipants,
    _downloads,
    _uploads: state.fs.uploads,
    progress: itemDetail.progress,
  }
}

const getEditingRows = (
  edits: I.Map<Types.EditID, Types.Edit>,
  parentPath: Types.Path
): Array<SortableEditingRowItem> =>
  edits
    .filter(edit => edit.parentPath === parentPath)
    .toArray()
    .map(([editID, edit]) => ({
      rowType: 'editing',
      editID,
      name: edit.name,
      // fields for sortable
      editType: edit.type,
      type: 'folder',
    }))

const getStillRows = (
  pathItems: I.Map<Types.Path, Types.PathItem>,
  parentPath: Types.Path,
  names: Array<string>
): Array<SortableStillRowItem> =>
  names.reduce((items, name) => {
    const item = pathItems.get(Types.pathConcat(parentPath, name), Constants.makeUnknownPathItem({name}))
    if (item.tlfMeta && item.tlfMeta.isIgnored) {
      return items
    }
    return [
      ...items,
      {
        rowType: 'still',
        path: Types.pathConcat(parentPath, item.name),
        name: item.name,
        // fields for sortable
        type: item.type,
        tlfMeta: item.tlfMeta,
        lastModifiedTimestamp: item.lastModifiedTimestamp,
      },
    ]
  }, [])

// TODO: when we have renames, reconcile editing rows in here too.
const amendStillRows = (
  stills: Array<SortableStillRowItem>,
  uploads: Types.Uploads
): Array<SortableRowItem> =>
  stills.map(still => {
    const {name, type, path} = still
    if (type === 'folder') {
      // Don't show an upload row for folders.
      return still
    }
    if (!uploads.writingToJournal.has(path) && !uploads.syncingPaths.has(path)) {
      // The entry is absent from uploads. So just show a still row.
      return still
    }
    return ({
      rowType: 'uploading',
      name,
      path,
      // field for sortable
      type,
    }: SortableUploadingRowItem)
  })

const getPlaceholderRows = type => [
  {rowType: 'placeholder', name: '1', type},
  {rowType: 'placeholder', name: '2', type},
  {rowType: 'placeholder', name: '3', type},
]

const getItemsFromStateProps = (stateProps, path: Types.Path) => {
  if (stateProps.progress === 'pending') {
    return getPlaceholderRows(Types.getPathLevel(path) <= 2 ? 'folder' : 'file')
  }

  const editingRows = getEditingRows(stateProps._edits, path)
  const stillRows = getStillRows(
    stateProps._pathItems,
    path,
    stateProps._itemChildren.union(stateProps._itemFavoriteChildren).toArray()
  )

  return sortRowItems(
    editingRows.concat(amendStillRows(stillRows, stateProps._uploads)),
    stateProps._sortSetting,
    Types.pathIsNonTeamTLFList(path) ? stateProps._username : undefined
  )
}

const mergeProps = (stateProps, dispatchProps, {path, routePath}) => ({
  isUserReset: stateProps.isUserReset,
  items: getItemsFromStateProps(stateProps, path),
  path,
  progress: stateProps.progress,
  resetParticipants: stateProps.resetParticipants,
  routePath,
})

export default compose(
  SecurityPrefsPromptingHoc,
  FilesLoadingHoc,
  connect(mapStateToProps, undefined, mergeProps),
  setDisplayName('Files')
)(Files)
